from setuptools import setup

setup(
    name="omnim",
    url="https://github.com/cet-t/omnipy",
    version="0.1.1",
    description="useful toolkit",
    author="cet",
    packages=["omnim"],
    include_package_data=True,
    zip_safe=True,
)
